@@ -49,6 +49,18 @@ struct _CslPcmDriver
 
 #define	ARTS_PCM_STREAM(pcmstream)	((ArtsPcmStream*) (pcmstream))
 
+struct _CslSampleDriver
+{
+  int n_samples;
+};
+
+typedef struct {
+  CslSample  sample;
+  int        id;
+} ArtsSample;
+
+#define	ARTS_SAMPLE(sample)		((ArtsSample*) (sample))
+
 /* --- prototypes --- */
 static CslErrorType  arts_pcm_driver_init		(CslDriver       *driver);
 static void	     arts_pcm_driver_shutdown		(CslDriver       *driver);
@@ -88,6 +100,34 @@ static CslErrorType  arts_pcm_set_channel_mapping       (CslPcmStream   *stream,
 							 const char     *mapping);
 static CslErrorType  pcm_device_arts_setup		(ArtsPcmStream	*stream);
 
+static CslErrorType	arts_sample_driver_init	        (CslDriver	*driver);
+static void		arts_sample_driver_shutdown     (CslDriver	*driver);
+
+static CslErrorType	arts_sample_new		        (CslDriver	*driver,
+					                 const char     *name,
+					                 const char     *role,
+					                 const char     *title,
+							 CslSample     **sample_p);
+static void		arts_sample_release		(CslSample	*sample);
+
+static int		arts_sample_write	        (CslSample	*sample,
+				 	                 unsigned int    n_bytes,
+				 	                 void           *bytes);
+static CslErrorType	arts_sample_write_done	        (CslSample	*sample);
+
+static CslErrorType	arts_sample_play		(CslSample	*sample);
+static CslBool		arts_sample_is_playing		(CslSample	*sample);
+static CslErrorType	arts_sample_stop		(CslSample	*sample);
+static CslErrorType	arts_sample_wait_done		(CslSample	*sample);
+
+static CslErrorType	arts_sample_cache_add	        (CslSample	*sample);
+static CslErrorType	arts_sample_cache_find	        (CslDriver      *driver,
+					                 const char     *name,
+					                 const char     *role,
+					                 const char     *title,
+							 CslSample     **sample_p);
+static void		arts_sample_cache_remove	(CslSample	*sample);
+
 
 /* --- varables --- */
 CslPcmDriverVTable _csl_arts_pcm_vtable = {
@@ -110,6 +150,22 @@ CslPcmDriverVTable _csl_arts_pcm_vtable = {
   arts_pcm_set_channel_mapping,
 };
 
+CslSampleDriverVTable _csl_arts_sample_vtable = {
+  arts_sample_driver_init,
+  arts_sample_driver_shutdown,
+  arts_sample_new,
+  arts_sample_release,
+  arts_sample_write,
+  arts_sample_write_done,
+  arts_sample_play,
+  arts_sample_is_playing,
+  arts_sample_stop,
+  arts_sample_wait_done,
+  arts_sample_cache_add,
+  arts_sample_cache_find,
+  arts_sample_cache_remove
+};
+
 
 /* --- functions --- */
 static void
@@ -252,7 +308,7 @@ arts_pcm_stream_init (CslDriver       *driver,
   ArtsPcmStream *stream;
   CslErrorType error;
   int stream_id;
-  int direction = readable ? ARTS_CSL_ADAPTER_DIR_INPUT : ARTS_CSL_ADAPTER_DIR_OUTPUT;
+  int direction = readable ? ARTS_CSL_DIR_INPUT : ARTS_CSL_DIR_OUTPUT;
   
   csl_return_val_if_fail ((writable && !readable) || (readable && !writable), CSL_EINTERN);
   
@@ -408,45 +464,29 @@ arts_pcm_write (CslPcmStream *_stream,
 static CslErrorType
 arts_pcm_update_status (CslPcmStream *_stream)
 {
-  /* FIXME */
-  return CSL_ENODRIVER;
-#if 0
+  CslPcmDriver *pcm_driver = _stream->driver->pcm_data;
   ArtsPcmStream *stream = ARTS_PCM_STREAM (_stream);
-  audio_buf_info info;
+  int bytes;
   
-  memset (&info, 0, sizeof (info));
+  ArtsCslStreamStatus *status = _arts_csl_adapter_status (pcm_driver->adapter, stream->id);
+
+  if (status == NULL) /* server crash? */
+    return CSL_EIO;
+
   if (stream->stream.readable)
-    (void) ioctl (stream->fd, SNDCTL_DSP_GETISPACE, &info);
-  else /* stream->stream.writable */
-    (void) ioctl (stream->fd, SNDCTL_DSP_GETOSPACE, &info);
-  
-  /* always fill package info, deduce stream buffer sizes from there */
-  stream->stream.packet.n_total_packets = info.fragstotal;
-  stream->stream.packet.packet_size = info.fragsize;
-  stream->stream.packet.n_packets_available = info.fragments;
-  stream->stream.packet.packet_watermark = 1; /* Arts supports just one packet */
-  
-  /* setup buffering values from packet information */
-  stream->stream.buffer_size = (stream->stream.packet.n_total_packets *
-				stream->stream.packet.packet_size);
-  stream->stream.n_bytes_available = info.bytes; /* info.fragments * info.fragsize */
-  stream->stream.buffer_watermark = (stream->stream.packet.packet_size *
-				     stream->stream.packet.packet_watermark);
-  /* kludge around buggy Arts drivers (e.g. es1371 in 2.3.34) */
-  stream->stream.n_bytes_available = MIN (stream->stream.n_bytes_available,
-					  stream->stream.buffer_size);
+    bytes = status->stream_buffer_used;
+  else
+    bytes = stream->stream.buffer_size - status->stream_buffer_used;
+
+  stream->stream.n_bytes_available = CLAMP (bytes, 0, stream->stream.buffer_size);
+  _arts_csl_stream_status_free (status);
+
   if (csl_debug (PCM))
-    csl_message ("Arts-STATUS(%s): left=%d/%d frags: total=%d size=%d count=%d bytes=%d",
-		 stream->device_name,
+    csl_message ("Arts-STATUS: buffer level=%d/%d bytes",
 		 stream->stream.n_bytes_available,
-		 stream->stream.buffer_size,
-		 info.fragstotal,
-		 info.fragsize,
-		 info.fragments,
-		 info.bytes);
+		 stream->stream.buffer_size);
   
   return CSL_ENONE;
-#endif
 }
 
 static CslErrorType
@@ -603,7 +643,7 @@ pcm_device_arts_setup (ArtsPcmStream *stream)
 {
   CslPcmDriver *pcm_driver = stream->stream.driver->pcm_data;
   
-  ArtsCslAdapterStreamParams params = {
+  ArtsCslStreamParams params = {
     stream->stream.rate,
     stream->stream.n_channels,
     stream->stream.format
@@ -725,5 +765,182 @@ pcm_device_arts_setup (ArtsPcmStream *stream)
 #endif
 }
 
+CslErrorType
+arts_sample_driver_init	(CslDriver *driver)
+{
+  /* FIXME: this assumes that the pcm driver will always be there, can I assume this here? */
+  if (driver->pcm_data)
+    {
+      /* initialize the driver's sample portion */
+      driver->sample_vtable = &_csl_arts_sample_vtable;
+      driver->sample_data = csl_new0 (CslSampleDriver, 1);
+      return CSL_ENONE;
+    }
+  else
+    return CSL_ENODRIVER;
+}
+
+void
+arts_sample_driver_shutdown (CslDriver *driver)
+{
+  csl_return_if_fail (driver != NULL);
+  csl_return_if_fail (driver->sample_data != NULL);
+
+  if (driver->sample_data->n_samples > 0)
+    csl_warning ("shutting down sample driver while %d samples still alive",
+		 driver->sample_data->n_samples);
+
+  csl_free (driver->sample_data);
+}
+
+CslErrorType
+arts_sample_new	(CslDriver	*driver,
+		 const char     *name,
+		 const char     *role,
+		 const char     *title,
+		 CslSample     **sample_p)
+{
+  CslPcmDriver *pcm_driver = driver->pcm_data;
+  int sample_id;
+  ArtsSample *sample;
+ 
+  sample_id = _arts_csl_adapter_sample_new (pcm_driver->adapter, name, role, title);
+  if (!sample_id)   /* server crash */
+    {
+      *sample_p = NULL;
+      return CSL_EIO;
+    }
+  
+  sample = csl_new0 (ArtsSample, 1);
+  sample->sample.driver = driver;
+  sample->id = sample_id;
+  *sample_p = &sample->sample;
+ 
+  driver->sample_data->n_samples++;
+  return CSL_ENONE;
+}
+
+void
+arts_sample_release (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_release (pcm_driver->adapter, sample->id);
+  sample->sample.driver->sample_data->n_samples--;
+
+  csl_free (_sample);
+}
+
+int
+arts_sample_write (CslSample	  *_sample,
+		   unsigned int    n_bytes,
+		   void           *bytes)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_write (pcm_driver->adapter, sample->id, bytes, n_bytes);
+  return n_bytes; /* FIXME: server crashes? */
+}
+
+CslErrorType
+arts_sample_write_done (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_write_done (pcm_driver->adapter, sample->id);
+  return CSL_ENONE; /* FIXME: server crashes? */
+}
+
+CslErrorType
+arts_sample_play (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+  CslBool result;
+
+  result = _arts_csl_adapter_sample_play (pcm_driver->adapter, sample->id);
+  return result ? CSL_ENONE : CSL_EIO; /* FIXME: is EIO right here? */
+}
+
+CslBool
+arts_sample_is_playing (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  return _arts_csl_adapter_sample_is_playing (pcm_driver->adapter, sample->id);
+}
+  
+CslErrorType
+arts_sample_stop (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_stop (pcm_driver->adapter, sample->id);
+  return CSL_ENONE; /* FIXME: server crashes */
+}
+  
+CslErrorType
+arts_sample_wait_done (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_wait_done (pcm_driver->adapter, sample->id);
+  return CSL_ENONE; /* FIXME: server crashes */
+}
+
+CslErrorType
+arts_sample_cache_add (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_cache_add (pcm_driver->adapter, sample->id);
+
+  return CSL_ENONE;
+}
+
+CslErrorType
+arts_sample_cache_find (CslDriver        *driver,
+			const char       *name,
+			const char       *role,
+			const char       *title,
+			CslSample       **sample_p)
+{
+  CslPcmDriver *pcm_driver = driver->pcm_data;
+  ArtsSample *sample;
+  int sample_id;
+  
+  sample_id = _arts_csl_adapter_sample_cache_find (pcm_driver->adapter, name, role, title);
+  if (!sample_id)
+    {
+      *sample_p = NULL;	/* server crash OR sample not found in cache */
+      return CSL_EIO;
+    }
+
+  sample = csl_new0 (ArtsSample, 1);
+  sample->sample.driver = driver;
+  sample->id = sample_id;
+  
+  driver->sample_data->n_samples++;
+  *sample_p = &sample->sample;
+
+  return CSL_ENONE;
+}
+
+void
+arts_sample_cache_remove (CslSample *_sample)
+{
+  ArtsSample *sample = ARTS_SAMPLE (_sample);
+  CslPcmDriver *pcm_driver = sample->sample.driver->pcm_data;
+
+  _arts_csl_adapter_sample_cache_remove (pcm_driver->adapter, sample->id);
+}
+
 /* vim:ts=8:sw=2:sts=2
  */
